# -*- coding: utf-8 -*-

import logging
import os
import sys
import tempfile
import unittest
from io import StringIO

import ldap3

import pytest

import lgmm

from .common import (
    LGMM_CONFIG,
    TestCase,
    logger,
)
from .data import (
    data,
    invalid_subjects,
)
from .mock_ldap import MockLIGOLDAP


class MapfileManagerTests(TestCase):
    def tearDown(self):
        # Stop tests polluting each other in Python2
        if "config" in sys.modules:
            del sys.modules["config"]

    def test_log_file_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.LOGFILE, "/var/log/lgmm/lgmm.log")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--log-file", "log-file-path"],
            logger=logger,
        )
        self.assertEqual(manager.LOGFILE, "log-file-path")

        with tempfile.NamedTemporaryFile() as logfile:
            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    LGMM_CONFIG,
                    "--log-file=%s" % (logfile.name,),
                    "--log-level=ERROR",
                ]
            )
        self.assertEqual(manager.LOGFILE, logfile.name)

    def test_log_file_config(self):
        with tempfile.NamedTemporaryFile() as logfile:
            with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
                with open(configfile.name, "w") as config:
                    config.write("LOGFILE='{}'".format(logfile.name))
                    config.flush()
                    manager = lgmm.MapfileManager(
                        ["lgmm", "--config", configfile.name, "--log-level=ERROR"]
                    )

                self.assertEqual(manager.LOGFILE, logfile.name)

            with open(configfile.name, "w") as config:
                config.write("LOGFILE='dummylogfile-name'")
                manager = lgmm.MapfileManager(
                    [
                        "lgmm",
                        "--config",
                        configfile.name,
                        "--log-file={}".format(logfile.name),
                        "--log-level=ERROR",
                    ]
                )

        self.assertEqual(manager.LOGFILE, logfile.name)

    def test_blacklist_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.BLACKLIST, None)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-b", "blackfile-path"],
            logger=logger,
        )
        self.assertEqual(manager.BLACKLIST, "blackfile-path")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--blacklist", "blackfile-path"],
            logger=logger,
        )
        self.assertEqual(manager.BLACKLIST, "blackfile-path")

    def test_blacklist_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("BLACKLIST='config-blackfile-path'")
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.BLACKLIST, "config-blackfile-path")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--blacklist", "blackfile-path"],
                logger=logger,
            )
            self.assertEqual(manager.BLACKLIST, "blackfile-path")

    @unittest.skip("Causes lgmm test script to daemonize")
    def test_daemon_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.DAEMON, False)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-d"], logger=logger
        )
        self.assertEqual(manager.DAEMON, True)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--daemon"], logger=logger
        )
        self.assertEqual(manager.DAEMON, True)

    def test_dry_run_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.DRY_RUN, False)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--dry-run"], logger=logger
        )
        self.assertEqual(manager.DRY_RUN, True)

    def test_dry_run_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("DRY_RUN=True")
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.DRY_RUN, True)

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--no-dry-run"], logger=logger
            )
            self.assertEqual(manager.DRY_RUN, False)

    def test_force_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.FORCE, False)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-f"], logger=logger
        )
        self.assertEqual(manager.FORCE, True)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--force"], logger=logger
        )
        self.assertEqual(manager.FORCE, True)

        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("FORCE=True")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.FORCE, True)

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--no-force"], logger=logger
            )
            self.assertEqual(manager.FORCE, False)

    def test_force_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("FORCE=True")
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.FORCE, True)

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--no-force"], logger=logger
            )
            self.assertEqual(manager.FORCE, False)

    def test_outfile_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.OUTFILE, "/etc/grid-security/grid-mapfile")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-g", "grid-mapfile-path"],
            logger=logger,
        )
        self.assertEqual(manager.OUTFILE, "grid-mapfile-path")

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--grid-mapfile",
                "grid-mapfile-path",
            ],
            logger=logger,
        )
        self.assertEqual(manager.OUTFILE, "grid-mapfile-path")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-o", "grid-mapfile-path"],
            logger=logger,
        )
        self.assertEqual(manager.OUTFILE, "grid-mapfile-path")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--outfile", "grid-mapfile-path"],
            logger=logger,
        )
        self.assertEqual(manager.OUTFILE, "grid-mapfile-path")

    def test_outfile_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("OUTFILE='outfile-path-config'")
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.OUTFILE, "outfile-path-config")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--outfile", "outfile-path"],
                logger=logger,
            )
            self.assertEqual(manager.OUTFILE, "outfile-path")

    def test_groups_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.GROUPS, [["Communities:LSCVirgoLIGOGroupMembers"]])

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--groups", "A,B,C"], logger=logger
        )
        self.assertEqual(manager.GROUPS, [["A", "B", "C"]])

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--groups",
                "A1,A2,A3",
                "--groups",
                "B1,B2",
                "--groups",
                "C1",
            ],
            logger=logger,
        )
        self.assertEqual(manager.GROUPS, [["A1", "A2", "A3"], ["B1", "B2"], ["C1"]])

    def test_groups_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('GROUPS=["A","B","C"]')
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.GROUPS, [["A", "B", "C"]])

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--groups", "D,E"], logger=logger
            )
            self.assertEqual(manager.GROUPS, [["D", "E"]])

    def test_infile_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.INFILE, "/etc/grid-security/grid-mapfile")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-i", "infile-path"], logger=logger
        )
        self.assertEqual(manager.INFILE, "infile-path")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--infile", "infile-path"],
            logger=logger,
        )
        self.assertEqual(manager.INFILE, "infile-path")

    def test_infile_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("INFILE='infile-path-config'")
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.INFILE, "infile-path-config")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--infile", "infile-path"],
                logger=logger,
            )
            self.assertEqual(manager.INFILE, "infile-path")

    def test_ldap_urls_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(
            manager.LDAP_URLS, ["ldaps://ldap.ligo.org", "ldaps://ldap.phys.uwm.edu"]
        )

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--ldap-urls", "A,B"], logger=logger
        )
        self.assertEqual(manager.LDAP_URLS, ["A", "B"])

    def test_ldap_urls_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('LDAP_URLS=["A", "B"]')
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.LDAP_URLS, ["A", "B"])

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--ldap-urls", "D,E"],
                logger=logger,
            )
            self.assertEqual(manager.LDAP_URLS, ["D", "E"])

    def test_ldap_base_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.LDAP_BASE, "dc=ligo,dc=org")

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--ldap-base",
                "ou=people,dc=ligo,dc=org",
            ],
            logger=logger,
        )
        self.assertEqual(manager.LDAP_BASE, "ou=people,dc=ligo,dc=org")

    def test_ldap_base_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('LDAP_BASE="ou=people,dc=ligo,dc=org"')
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.LDAP_BASE, "ou=people,dc=ligo,dc=org")

            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    configfile.name,
                    "--ldap-base",
                    "ou=robot,dc=ligo,dc=org",
                ],
                logger=logger,
            )
            self.assertEqual(manager.LDAP_BASE, "ou=robot,dc=ligo,dc=org")

    def test_ldap_scope_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.LDAP_SCOPE, "subtree")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--ldap-scope", "onelevel"],
            logger=logger,
        )
        self.assertEqual(manager.LDAP_SCOPE, "onelevel")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--ldap-scope", "one"], logger=logger
        )
        self.assertEqual(manager.LDAP_SCOPE, "onelevel")

        with self.assertRaises(RuntimeError):
            manager = lgmm.MapfileManager(
                ["lgmm", "--ldap-scope", "noscope"], logger=logger
            )

    def test_ldap_scope_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('LDAP_SCOPE="onelevel"')
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.LDAP_SCOPE, "onelevel")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--ldap-scope", "base"],
                logger=logger,
            )
            self.assertEqual(manager.LDAP_SCOPE, "base")

        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('LDAP_SCOPE="noscope"')

            with self.assertRaises(RuntimeError):
                manager = lgmm.MapfileManager(
                    ["lgmm", "--config", configfile.name], logger=logger
                )

    def test_log_level_argument(self):
        with self.assertRaises(AttributeError):
            manager = lgmm.MapfileManager(
                ["lgmm", "--log-level", "mylevel"], logger=logger
            )

        with tempfile.NamedTemporaryFile() as logfile:
            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    LGMM_CONFIG,
                    "--log-file=%s" % (logfile.name,),
                    "--log-level=ERROR",
                ]
            )
        self.assertEqual(manager.logger.level, logging.ERROR)

        with tempfile.NamedTemporaryFile() as logfile:
            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    LGMM_CONFIG,
                    "--log-file=%s" % (logfile.name,),
                    "--log-level=CRITICAL",
                ]
            )
        self.assertEqual(manager.logger.level, logging.CRITICAL)

    def test_log_level_config(self):
        with tempfile.NamedTemporaryFile() as logfile:
            with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
                with open(configfile.name, "w") as config:
                    config.write("LOG_LEVEL='CRITICAL'")
                    config.flush()
                    manager = lgmm.MapfileManager(
                        [
                            "lgmm",
                            "--config",
                            configfile.name,
                            "--log-file={}".format(logfile.name),
                        ]
                    )

                self.assertEqual(manager.logger.level, logging.CRITICAL)

                with open(configfile.name, "w") as config:
                    config.write("import logging; LOG_LEVEL = logging.CRITICAL")
                    config.flush()
                    manager = lgmm.MapfileManager(
                        [
                            "lgmm",
                            "--config",
                            configfile.name,
                            "--log-file={}".format(logfile.name),
                        ]
                    )

                self.assertEqual(manager.logger.level, logging.CRITICAL)

                with open(configfile.name, "w") as config:
                    config.write("LOG_LEVEL=logging.CRITICAL")
                    manager = lgmm.MapfileManager(
                        [
                            "lgmm",
                            "--config",
                            configfile.name,
                            "--log-file={}".format(logfile.name),
                            "--log-level=ERROR",
                        ]
                    )

                self.assertEqual(manager.logger.level, logging.ERROR)

    def test_member_attr_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.MEMBER_ATTR, "isMemberOf")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-m", "groupMembership"],
            logger=logger,
        )
        self.assertEqual(manager.MEMBER_ATTR, "groupMembership")

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--membership-attribute",
                "groupMembership",
            ],
            logger=logger,
        )
        self.assertEqual(manager.MEMBER_ATTR, "groupMembership")

    def test_member_attr_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('MEMBER_ATTR="isMember"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.MEMBER_ATTR, "isMember")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--membership-attr", "member"],
                logger=logger,
            )
            self.assertEqual(manager.MEMBER_ATTR, "member")

    def test_login_attr_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.LOGIN_ATTR, "x-LIGO-uid")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-L", "uid"], logger=logger
        )
        self.assertEqual(manager.LOGIN_ATTR, "uid")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--login-attribute", "uid"],
            logger=logger,
        )
        self.assertEqual(manager.LOGIN_ATTR, "uid")

    def test_login_attr_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('LOGIN_ATTR="uid"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.LOGIN_ATTR, "uid")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--login-attr", "posixUid"],
                logger=logger,
            )
            self.assertEqual(manager.LOGIN_ATTR, "posixUid")

    def test_no_local_accounts_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.NO_LOCAL_ACCOUNTS, False)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--no-local-accounts"], logger=logger
        )
        self.assertEqual(manager.NO_LOCAL_ACCOUNTS, True)

        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("NO_LOCAL_ACCOUNTS=True")
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.NO_LOCAL_ACCOUNTS, True)

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--local-accounts"], logger=logger
            )
            self.assertEqual(manager.NO_LOCAL_ACCOUNTS, False)

    def test_no_local_accounts_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("NO_LOCAL_ACCOUNTS=False")
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.NO_LOCAL_ACCOUNTS, False)

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--no-local-accounts"],
                logger=logger,
            )
            self.assertEqual(manager.NO_LOCAL_ACCOUNTS, True)

    def test_pidfile_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.PIDFILE, "/var/run/lgmm/lgmm.pid")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--pid-file", "pidfile-path"],
            logger=logger,
        )
        self.assertEqual(manager.PIDFILE, "pidfile-path")

    def test_pidfile_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('PIDFILE="pidfile-path-config"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.PIDFILE, "pidfile-path-config")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--pid-file", "pidfle-path"],
                logger=logger,
            )
            self.assertEqual(manager.PIDFILE, "pidfle-path")

    def test_site_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.SITE, None)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-s", "cdf"], logger=logger
        )
        self.assertEqual(manager.SITE, "cdf")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--site", "cdf"], logger=logger
        )
        self.assertEqual(manager.SITE, "cdf")

    def test_site_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('SITE="cdf"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.SITE, "cdf")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--site", "cit"], logger=logger
            )
            self.assertEqual(manager.SITE, "cit")

    def test_site_attr_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.SITE_ATTR, "x-ligo-ldg-{site}")

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--site-attr",
                "x-private-{site}-attr",
            ],
            logger=logger,
        )
        self.assertEqual(manager.SITE_ATTR, "x-private-{site}-attr")

    def test_site_attr_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('SITE_ATTR="x-private-{site}-attr"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.SITE_ATTR, "x-private-{site}-attr")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--site-attr", "x-priv-{site}"],
                logger=logger,
            )
            self.assertEqual(manager.SITE_ATTR, "x-priv-{site}")

    def test_sleep_time_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.SLEEP_TIME, 300)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--sleep-time", "600"], logger=logger
        )
        self.assertEqual(manager.SLEEP_TIME, 600)

    def test_sleep_time_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("SLEEP_TIME=500")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.SLEEP_TIME, 500)

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--sleep-time", "600"],
                logger=logger,
            )
            self.assertEqual(manager.SLEEP_TIME, 600)

    def test_sort_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.SORT, None)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--sort", "ABC"], logger=logger
        )
        self.assertEqual(manager.SORT, "ABC")

    def test_sort_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('SORT="ABC"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.SORT, "ABC")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--sort", "DEF"], logger=logger
            )
            self.assertEqual(manager.SORT, "DEF")

    def test_random_delay_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.RANDOM_DELAY, True)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--random-delay"], logger=logger
        )
        self.assertEqual(manager.RANDOM_DELAY, True)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--no-random-delay"], logger=logger
        )
        self.assertEqual(manager.RANDOM_DELAY, False)

    def test_random_delay_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("RANDOM_DELAY=True")
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.RANDOM_DELAY, True)

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--no-random-delay"],
                logger=logger,
            )
            self.assertEqual(manager.RANDOM_DELAY, False)

    def test_user_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.USER, "nobody")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--user", "gridmap"], logger=logger
        )
        self.assertEqual(manager.USER, "gridmap")

    def test_user_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('USER="gridmap"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.USER, "gridmap")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--user", "nobody"], logger=logger
            )
            self.assertEqual(manager.USER, "nobody")

    def test_whitelist_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.WHITELIST, None)

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "-w", "whitelist-path"],
            logger=logger,
        )
        self.assertEqual(manager.WHITELIST, "whitelist-path")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--whitelist", "whitelist-path"],
            logger=logger,
        )
        self.assertEqual(manager.WHITELIST, "whitelist-path")

    def test_whitelist_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('WHITELIST="whitelist-path"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.WHITELIST, "whitelist-path")

            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    configfile.name,
                    "--whitelist",
                    "alt-whitelist-path",
                ],
                logger=logger,
            )
            self.assertEqual(manager.WHITELIST, "alt-whitelist-path")

    def test_filter_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.FILTER, None)

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--filter",
                "(!(loginShell=/bin/nologin))",
            ],
            logger=logger,
        )
        self.assertEqual(manager.FILTER, "(!(loginShell=/bin/nologin))")

    def test_filter_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('FILTER="(!(loginShell=/bin/nologin))"')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.FILTER, "(!(loginShell=/bin/nologin))")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--filter", "(!(active=False))"],
                logger=logger,
            )
            self.assertEqual(manager.FILTER, "(!(active=False))")

    def test_login_whitelist_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.LOGIN_WHITELIST, None)

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--login-whitelist",
                "Albert,Marie,Paul",
            ],
            logger=logger,
        )
        self.assertEqual(manager.LOGIN_WHITELIST, ["Albert", "Marie", "Paul"])

    def test_login_whitelist_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('LOGIN_WHITELIST=["A","B","C"]')

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.LOGIN_WHITELIST, ["A", "B", "C"])

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--login-whitelist", "D,E"],
                logger=logger,
            )
            self.assertEqual(manager.LOGIN_WHITELIST, ["D", "E"])

    def test_username_policy_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.USERNAME_POLICY, "SITE_ONLY")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--username-policy", "SITE_FIRST"],
            logger=logger,
        )

        self.assertEqual(manager.USERNAME_POLICY, "SITE_FIRST")

    def test_username_policy_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("USERNAME_POLICY='GLOBAL_ONLY'")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.USERNAME_POLICY, "GLOBAL_ONLY")

            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    configfile.name,
                    "--username-policy",
                    "GLOBAL_FIRST",
                ],
                logger=logger,
            )
            self.assertEqual(manager.USERNAME_POLICY, "GLOBAL_FIRST")

    def test_login_groups_map_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.LOGIN_GROUPS_MAP, None)

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--login-groups-map",
                "cbc:LDGCBCGroup",
            ],
            logger=logger,
        )

        self.assertEqual(manager.LOGIN_GROUPS_MAP, [(["cbc"], ["LDGCBCGroup"])])

    def test_login_groups_map_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("LOGIN_GROUPS_MAP=[('cbc', ['LDGCBCGroup'])]")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.LOGIN_GROUPS_MAP, [("cbc", ["LDGCBCGroup"])])

            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    configfile.name,
                    "--login-groups-map",
                    "burst:LDGBurstGroup",
                ],
                logger=logger,
            )
            self.assertEqual(manager.LOGIN_GROUPS_MAP, [(["burst"], ["LDGBurstGroup"])])

    def test_subject_transform_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.SUBJECT_TRANSFORM, "AUTO")

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--subject-transform", "AUTO"],
            logger=logger,
        )

        self.assertEqual(manager.SUBJECT_TRANSFORM, "AUTO")

    def test_subject_transform_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("SUBJECT_TRANSFORM='NONE'")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.SUBJECT_TRANSFORM, "NONE")

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--subject-transform", "AUTO"],
                logger=logger,
            )
            self.assertEqual(manager.SUBJECT_TRANSFORM, "AUTO")

    def test_overrides_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.OVERRIDES, None)

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--overrides",
                """{
                    "conditions": {"eduPersonPrincipalName": []},
                    "logins_append": ["{eduPersonPrincipalName}"],
                }""",
            ],
            logger=logger,
        )

        self.assertEqual(
            manager.OVERRIDES,
            [
                {
                    "conditions": {"eduPersonPrincipalName": []},
                    "logins_append": ["{eduPersonPrincipalName}"],
                }
            ],
        )

    def test_overrides_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write(
                    """OVERRIDES=[{
                    "conditions": {"eduPersonPrincipalName": []},
                    "logins_append": ["{eduPersonPrincipalName}"],
                }]"""
                )

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(
                manager.OVERRIDES,
                [
                    {
                        "conditions": {"eduPersonPrincipalName": []},
                        "logins_append": ["{eduPersonPrincipalName}"],
                    }
                ],
            )

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--overrides", "[{}]"],
                logger=logger,
            )
            self.assertEqual(manager.OVERRIDES, [{}])

    def test_version_argument(self):
        capturedOutput = StringIO()
        sys.stdout = capturedOutput
        with self.assertRaises(SystemExit):
            lgmm.MapfileManager(
                ["lgmm", "--config", LGMM_CONFIG, "--version"], logger=logger
            )

        sys.stdout = sys.__stdout__
        value = capturedOutput.getvalue()
        self.assertRegex(value, r"^lgmm [0-9]+.[0-9]+.[0-9]+\+?$")

    def test_help_argument(self):
        capturedOutput = StringIO()
        sys.stdout = capturedOutput
        with self.assertRaises(SystemExit):
            lgmm.MapfileManager(["lgmm", "--help"], logger=logger)

        sys.stdout = sys.__stdout__
        value = capturedOutput.getvalue()

        self.assertIn("Gridmapfile Input Options:", value)
        self.assertIn("Gridmapfile Output Options:", value)
        self.assertIn("Logging Options:", value)
        self.assertIn("Daemon Options:", value)
        self.assertIn("LDAP Host Options:", value)
        self.assertIn("LDAP Search Options:", value)
        self.assertIn("Additional Processing Options:", value)

    def test_dump_vars_argument(self):
        capturedOutput = StringIO()
        sys.stdout = capturedOutput
        with self.assertRaises(SystemExit):
            lgmm.MapfileManager(
                ["lgmm", "--config", LGMM_CONFIG, "--dump-vars"], logger=logger
            )

        sys.stdout = sys.__stdout__
        value = capturedOutput.getvalue()

        self.assertIn("BLACKLIST = None", value)
        self.assertIn("DAEMON = False", value)
        self.assertIn("DRY_RUN = False", value)
        self.assertIn("FILTER = None", value)
        self.assertIn("FORCE = False", value)
        self.assertIn("GROUPS = [['Communities:LSCVirgoLIGOGroupMembers']]", value)
        self.assertIn("INFILE = '/etc/grid-security/grid-mapfile'", value)
        self.assertIn("LDAP_BASE = 'dc=ligo,dc=org'", value)
        self.assertIn("LDAP_SCOPE = 'subtree'", value)
        self.assertIn(
            "LDAP_URLS = ['ldaps://ldap.ligo.org', 'ldaps://ldap.phys.uwm.edu']", value
        )
        self.assertIn("LOGFILE = '/var/log/lgmm/lgmm.log'", value)
        self.assertIn("LOGIN_ATTR = 'x-LIGO-uid'", value)
        self.assertIn("LOGIN_GROUPS_MAP = None", value)
        self.assertIn("LOGIN_WHITELIST = None", value)
        self.assertIn("LOG_LEVEL = 'INFO'", value)
        self.assertIn("MEMBER_ATTR = 'isMemberOf'", value)
        self.assertIn("NO_LOCAL_ACCOUNTS = False", value)
        self.assertIn("OUTFILE = '/etc/grid-security/grid-mapfile'", value)
        self.assertIn("OVERRIDES = None", value)
        self.assertIn("PIDFILE = '/var/run/lgmm/lgmm.pid'", value)
        self.assertIn("RANDOM_DELAY = True", value)
        self.assertIn("SITE = None", value)
        self.assertIn("SLEEP_TIME = 300", value)
        self.assertIn("SORT = None", value)
        self.assertIn("USER = 'nobody'", value)
        self.assertIn("USERNAME_POLICY = 'SITE_ONLY'", value)
        self.assertIn("WHITELIST = None", value)

    def test_remove_pid(self):
        with tempfile.NamedTemporaryFile(delete=False) as pidfile:
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", LGMM_CONFIG, "--pid-file", pidfile.name],
                logger=logger,
            )

            with open(pidfile.name, "w") as f:
                f.write("\n")

            self.assertTrue(os.path.exists(pidfile.name))

            manager.remove_pid()

            try:
                self.assertFalse(os.path.exists(pidfile.name))
            except AssertionError:
                pidfile.close()


class MapfileManagerFoldNewMappingTests(unittest.TestCase):
    def setUp(self):
        self.existing = lgmm.Mapping()
        self.white = lgmm.Mapping()
        self.black = lgmm.Mapping()
        self.central = lgmm.Mapping()

        self.manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )

    def fold_new_mapping(self):
        return self.manager.fold_new_mapping(
            self.existing, self.white, self.black, self.central
        )

    def test_no_mappings(self):
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 0)

    def test_only_existing_mappings(self):
        self.existing[data["einstein_dn"]] = data["einstein_usernames"]
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 0)

    def test_whitelist_mappings(self):
        self.white[data["einstein_dn"]] = data["einstein_usernames"]
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 1)
        self.assertEqual(new[data["einstein_dn"]], data["einstein_usernames"])

    def test_central_mappings_no_existing_no_force(self):
        self.manager.FORCE = False
        self.central[data["einstein_dn"]] = data["einstein_usernames"]

        new = self.fold_new_mapping()

        self.assertEqual(len(new), 0)

    def test_central_mappings_existing_no_force(self):
        self.manager.FORCE = False
        self.existing[data["einstein_dn"]] = data["einstein_usernames"]
        self.central[data["einstein_dn"]] = data["einstein_usernames"]

        new = self.fold_new_mapping()

        self.assertEqual(len(new), 1)
        self.assertEqual(new[data["einstein_dn"]], data["einstein_usernames"])

    def test_central_mappings_no_existing_force(self):
        self.manager.FORCE = True
        self.central[data["einstein_dn"]] = data["einstein_usernames"]

        new = self.fold_new_mapping()

        self.assertEqual(len(new), 1)
        self.assertEqual(new[data["einstein_dn"]], data["einstein_usernames"])

    def test_blacklist_mappings(self):
        self.black[data["einstein_dn"]] = data["einstein_usernames"]
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 0)

        self.central[data["einstein_dn"]] = data["einstein_usernames"]
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 0)

        self.manager.force = True
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 0)

    def test_blacklist_mappings_username_only(self):
        self.black[0] = data["einstein_usernames"]
        self.central[data["einstein_dn"]] = data["einstein_usernames"]
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 0)

        self.black[0] = ["albert"]
        self.central[data["einstein_dn"]] = data["einstein_usernames"]
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 0)

    def test_blacklist_and_whitelist_mappings(self):
        self.black[data["einstein_dn"]] = data["einstein_usernames"]
        self.white[data["einstein_dn"]] = data["einstein_usernames"]
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 1)
        self.central[data["einstein_dn"]] = data["einstein_usernames"]

    def test_blacklist_username_only_and_whitelist_mappings(self):
        self.black[0] = data["einstein_usernames"]
        self.white[data["einstein_dn"]] = data["einstein_usernames"]
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 1)
        self.central[data["einstein_dn"]] = data["einstein_usernames"]

        self.black[0] = "albert"
        new = self.fold_new_mapping()

        self.assertEqual(len(new), 1)
        self.central[data["einstein_dn"]] = ["albert.einstein"]

    def test_invalid_subject_strings(self):
        self.manager.FORCE = True
        for subject in invalid_subjects:
            self.central[subject] = data["einstein_usernames"]

            new = self.fold_new_mapping()

            self.assertEqual(len(new), 0)


@pytest.mark.skipif(
    ldap3.__version__ < "2.5.0",
    reason="ldap3 too old (< 2.5.0)",
)
class MapfileManagerIntegrationTests(MockLIGOLDAP):
    def setUp(self):
        super(MapfileManagerIntegrationTests, self).setUp()
        self.manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )

    def test_integration(self):
        self.manager.FORCE = True
        with tempfile.NamedTemporaryFile() as outfile:
            self.manager.OUTFILE = outfile.name

            self.manager.run()

            with open(outfile.name, "r") as gridmap:
                output = gridmap.read().splitlines()

        self.assertEqual(len(output), 3)
        self.assertIn(data["einstein_mapfile_entry"], output)
        self.assertIn(data["curie_mapfile_entry"], output)
        self.assertIn(data["dirac_mapfile_entry"], output)
