# -*- coding: utf-8 -*-

import sys
import tempfile

import lgmm

from .common import (
    LGMM_CONFIG,
    TestCase,
    logger,
)


# lgmm requires that each source have at least one non-default value set
def extra_args(name, sources):
    if name == "LDAP_URLS":
        argument = "ldap-scope"
        value = "onelevel"
    else:
        argument = "ldap-urls"
        value = "ldaps://ldap.gwastronomy.org"

    args = []
    for source in sources:
        args.extend(["--{}_{}".format(argument, source), value])
    return args


class MapfileManagerTests(TestCase):
    def tearDown(self):
        # Stop tests polluting each other in Python2
        if "config" in sys.modules:
            del sys.modules["config"]

    def test_sources_argument(self):
        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG], logger=logger
        )
        self.assertEqual(manager.SOURCES, ["default"])

        # lgmm requires that each source have at least one non-default value set
        with self.assertRaises(RuntimeError):
            manager = lgmm.MapfileManager(
                ["lgmm", "--sources", "default,gwastronomy"], logger=logger
            )

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--sources",
                "default,gwastronomy",
                "--ldap-urls_gwastronomy",
                "ldaps://ldap.gwastronomy.org",
            ],
            logger=logger,
        )

        self.assertEqual(manager.SOURCES, ["default", "gwastronomy"])

    def test_sources_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('SOURCES=["default", "gwastronomy"]')

            # lgmm requires that each source have at least one non-default value set
            with self.assertRaises(RuntimeError):
                manager = lgmm.MapfileManager(
                    ["lgmm", "--config", configfile.name], logger=logger
                )

            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    configfile.name,
                    "--ldap-urls_gwastronomy",
                    "ldaps://ldap.gwastronomy.org",
                ],
                logger=logger,
            )

            self.assertEqual(manager.SOURCES, ["default", "gwastronomy"])

            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    configfile.name,
                    "--sources",
                    "ligo,kagra",
                    "--ldap-urls_ligo",
                    "ldaps://ldap.gwastronomy.org",
                    "--ldap-urls_kagra",
                    "ldaps://ldap.gwastronomy.org",
                ],
                logger=logger,
            )
            self.assertEqual(manager.SOURCES, ["ligo", "kagra"])

    def test_groups_argument_custom(self):
        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--sources",
                "default,B",
                "--ldap-urls_B",
                "ldaps://ldap.b.com",
            ],
            logger=logger,
        )
        self.assertEqual(
            manager.get_config("GROUPS"), [["Communities:LSCVirgoLIGOGroupMembers"]]
        )
        self.assertEqual(
            manager.get_config("GROUPS", "default"),
            [["Communities:LSCVirgoLIGOGroupMembers"]],
        )
        self.assertEqual(
            manager.get_config("GROUPS", "B"),
            [["Communities:LSCVirgoLIGOGroupMembers"]],
        )

        # Request config for a source that does not exist, should it generate a failure or warning?
        # lgmm requires that each source have at least one non-default value set
        with self.assertRaises(RuntimeError):
            manager.get_config("GROUPS", "A")

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--sources",
                "default,B",
                "--ldap-urls_B",
                "ldaps://ldap.b.com",
                "--groups",
                "A,B,C",
            ],
            logger=logger,
        )
        self.assertEqual(manager.get_config("GROUPS", "default"), [["A", "B", "C"]])
        self.assertEqual(manager.get_config("GROUPS", "B"), [["A", "B", "C"]])

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--sources",
                "default,B",
                "--groups_B",
                "A,B,C",
            ],
            logger=logger,
        )
        self.assertEqual(
            manager.get_config("GROUPS", "default"),
            [["Communities:LSCVirgoLIGOGroupMembers"]],
        )
        self.assertEqual(manager.get_config("GROUPS", "B"), [["A", "B", "C"]])

        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--groups",
                "A1,A2,A3",
                "--groups",
                "B1,B2",
                "--groups",
                "C1",
            ],
            logger=logger,
        )
        self.assertEqual(manager.GROUPS, [["A1", "A2", "A3"], ["B1", "B2"], ["C1"]])

    def test_groups_config(self):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write('GROUPS=["A","B","C"]')
            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name], logger=logger
            )

            self.assertEqual(manager.GROUPS, [["A", "B", "C"]])

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, "--groups", "D,E"], logger=logger
            )
            self.assertEqual(manager.GROUPS, [["D", "E"]])

    def test_ldap_urls_argument(self):
        self._test(
            "LDAP_URLS",
            "-U",
            "--ldap-urls",
            ["ldaps://ldap.ligo.org", "ldaps://ldap.phys.uwm.edu"],
            ["A", "B"],
        )

    def test_ldap_base_argument(self):
        self._test(
            "LDAP_BASE",
            "-B",
            "--ldap-base",
            "dc=ligo,dc=org",
            "ou=people,dc=ligo,dc=org",
        )

    def test_ldap_scope_argument(self):
        self._test("LDAP_SCOPE", "-S", "--ldap-scope", "subtree", "onelevel")

    def test_site_attr_argument(self):
        self._test(
            "SITE_ATTR", [], "--site-attr", "x-ligo-ldg-{site}", "x-private-{site}-attr"
        )

    def test_groups_argument(self):
        self._test(
            "GROUPS",
            [],
            "--groups",
            [["Communities:LSCVirgoLIGOGroupMembers"]],
            [
                [
                    "Communities:LVC:LSC:LDG:CDF:LDGCDFUsers",
                    "Communities:LSCVirgoLIGOGroupMembers",
                ]
            ],
            "Communities:LVC:LSC:LDG:CDF:LDGCDFUsers,Communities:LSCVirgoLIGOGroupMembers",
        )

    def test_membership_attr_argument(self):
        self._test(
            "MEMBER_ATTR",
            ["-m", "-M"],
            "--membership-attribute",
            "isMemberOf",
            "member",
        )

    def test_login_attr_argument(self):
        self._test("LOGIN_ATTR", "-L", "--login-attribute", "x-LIGO-uid", "uid")

    def test_subject_attr_argument(self):
        self._test(
            "SUBJECT_ATTR", [], "--subject-attribute", "gridX509subject", "gridSubject"
        )

    def test_filter_argument(self):
        self._test("FILTER", [], "--filter", None, "givenName=Paul")

    def test_site_argument(self):
        self._test("SITE", [], "--site", None, "cdf")

    def test_login_whitelist_argument(self):
        self._test("LOGIN_WHITELIST", [], "--login-whitelist", None, ["Paul", "Steve"])

    def test_username_policy_argument(self):
        self._test(
            "USERNAME_POLICY", [], "--username-policy", "SITE_ONLY", "GLOBAL_ONLY"
        )

    def test_login_groups_map_argument(self):
        self._test(
            "LOGIN_GROUPS_MAP",
            [],
            "--login-groups-map",
            None,
            [(["cbc"], ["CBCGroup"])],
            "cbc:CBCGroup",
        )

    def test_subject_transform_argument(self):
        self._test("SUBJECT_TRANSFORM", [], "--subject-transform", "AUTO", "NONE")

    def test_overrides_argument(self):
        self._test(
            "OVERRIDES",
            "-O",
            "--overrides",
            None,
            [
                {
                    "conditions": {"eduPersonPrincipalName": []},
                    "logins_append": ["{eduPersonPrincipalName}"],
                }
            ],
            '{"conditions": {"eduPersonPrincipalName": []}, "logins_append": ["{eduPersonPrincipalName}"],}',
        )

    def _test(
        self, name, sarguments, larguments, default_value, new_value, newvalue_args=None
    ):
        for source_list in [
            ["default"],
            ["kagra", "lisc"],
            ["default", "kagra", "lisc"],
        ]:
            self._test_with_sources(
                name,
                sarguments,
                larguments,
                default_value,
                new_value,
                newvalue_args,
                source_list,
            )

    def _test_with_sources(
        self,
        name,
        sarguments,
        larguments,
        default_value,
        new_value,
        newvalue_args=None,
        sources=["default"],
    ):
        sarguments = lgmm.as_list(sarguments)
        larguments = lgmm.as_list(larguments)

        if newvalue_args is None:
            newvalue_args = ",".join(lgmm.as_list(new_value))

        self._test_argument_default(name, default_value, sources)

        for argument in sarguments + larguments:
            self._test_argument_newdefault(
                name, argument, new_value, newvalue_args, sources
            )
            self._test_config_newdefault(name, argument, new_value, sources)

            self._test_config_then_argument_newdefault(
                name, argument, new_value, newvalue_args, sources
            )

        self._test_config_source_specific(name, default_value, new_value, sources)

        for argument in larguments:
            self._test_argument_newdefault_altsyntax(
                name, argument, new_value, newvalue_args, sources
            )

            self._test_argument_source_specific(
                name, argument, default_value, new_value, newvalue_args, sources
            )

            self._test_config_then_argument_source_specific(
                name, argument, default_value, new_value, newvalue_args, sources
            )

    def _test_argument_default(self, name, default_value, sources):
        # Delete old config before running this test
        self.tearDown()

        manager = lgmm.MapfileManager(
            ["lgmm", "--config", LGMM_CONFIG, "--sources", ",".join(sources)]
            + extra_args(name, sources),
            logger=logger,
        )

        # Test default source gets default value
        self.assertEqual(manager.get_config(name), default_value)

        # Check all sources get default value
        for source in sources:
            self.assertEqual(manager.get_config(name, source), default_value)

    def _test_argument_newdefault(
        self, name, argument, new_value, newvalue_args, sources
    ):
        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--sources",
                ",".join(sources),
                argument,
                newvalue_args,
            ]
            + extra_args(name, sources),
            logger=logger,
        )
        # test default source gets new default
        self.assertEqual(manager.get_config(name), new_value)
        self.assertEqual(manager.get_config(name, "default"), new_value)

        # test all sources get new default
        for source in sources:
            self.assertEqual(manager.get_config(name, source), new_value)

    def _test_argument_source_specific(
        self, name, argument, default_value, specific_value, specificvalue_args, sources
    ):
        # Delete old config before running this test
        self.tearDown()

        # test setting non-default argument
        for source in [s for s in sources if s != "default"]:
            manager = lgmm.MapfileManager(
                [
                    "lgmm",
                    "--config",
                    LGMM_CONFIG,
                    "--sources",
                    ",".join(sources),
                    "_".join([argument, source]),
                    specificvalue_args,
                ]
                + extra_args(name, sources),
                logger=logger,
            )

            # test default source gets defualt value
            self.assertEqual(manager.get_config(name), default_value)
            self.assertEqual(manager.get_config(name, "default"), default_value)

            # test this source gets specific value
            self.assertEqual(manager.get_config(name, source), specific_value)

            # test other sources get the default value
            for other_source in [s for s in sources if s != source]:
                self.assertEqual(manager.get_config(name, other_source), default_value)

    def _test_argument_newdefault_altsyntax(
        self, name, argument, new_value, newvalue_args, sources
    ):
        # Test that we can use the syntax "--argument_default=X" to set the default value
        # This is equivalent to "--argument=X"
        manager = lgmm.MapfileManager(
            [
                "lgmm",
                "--config",
                LGMM_CONFIG,
                "--sources",
                ",".join(sources),
                "_".join([argument, "default"]),
                newvalue_args,
            ]
            + extra_args(name, sources),
            logger=logger,
        )

        # test default source gets defualt value
        self.assertEqual(manager.get_config(name), new_value)
        self.assertEqual(manager.get_config(name, "default"), new_value)

        for source in sources:
            # test this source gets specific value
            self.assertEqual(manager.get_config(name, source), new_value)

    def _test_config_newdefault(self, name, argument, new_value, sources):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("SOURCES={}\n".format(sources))
                config.write("{}={}\n".format(name, repr(new_value)))

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name] + extra_args(name, sources),
                logger=logger,
            )

        # test default source gets new default
        self.assertEqual(manager.get_config(name), new_value)
        self.assertEqual(manager.get_config(name, "default"), new_value)

        # test all sources get new default
        for source in sources:
            self.assertEqual(manager.get_config(name, source), new_value)

    def _test_config_then_argument_newdefault(
        self, name, argument, new_value, newvalue_args, sources
    ):
        with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
            with open(configfile.name, "w") as config:
                config.write("SOURCES={}\n".format(sources))
                config.write("{}={}\n".format(name, "'DUMMY'"))

            manager = lgmm.MapfileManager(
                ["lgmm", "--config", configfile.name, argument, newvalue_args]
                + extra_args(name, sources),
                logger=logger,
            )

        # test default source gets new default
        self.assertEqual(manager.get_config(name), new_value)
        self.assertEqual(manager.get_config(name, "default"), new_value)

        # test all sources get new default
        for source in sources:
            self.assertEqual(manager.get_config(name, source), new_value)

    def _test_config_source_specific(
        self, name, default_value, specific_value, sources
    ):
        # test setting non-default argument
        for source in [s for s in sources if s != "default"]:
            # Delete old config before each loop
            self.tearDown()

            with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
                with open(configfile.name, "w") as config:
                    config.write("SOURCES={}\n".format(sources))
                    config.write(
                        "{}_{}={}\n".format(name, source, repr(specific_value))
                    )

                manager = lgmm.MapfileManager(
                    ["lgmm", "--config", configfile.name] + extra_args(name, sources),
                    logger=logger,
                )

            # test default source gets defualt value
            self.assertEqual(manager.get_config(name), default_value)
            self.assertEqual(manager.get_config(name, "default"), default_value)

            # test this source gets specific value
            self.assertEqual(manager.get_config(name, source), specific_value)

            # test other sources get the default value
            for other_source in [s for s in sources if s != source]:
                self.assertEqual(manager.get_config(name, other_source), default_value)

    def _test_config_then_argument_source_specific(
        self, name, argument, default_value, specific_value, specificvalue_args, sources
    ):
        # test setting non-default argument
        for source in [s for s in sources if s != "default"]:
            # Delete old config before each loop
            self.tearDown()

            with tempfile.NamedTemporaryFile(suffix=".py") as configfile:
                with open(configfile.name, "w") as config:
                    config.write("SOURCES={}\n".format(sources))
                    config.write("{}_{}={}\n".format(name, source, "'DUMMY'"))

                manager = lgmm.MapfileManager(
                    [
                        "lgmm",
                        "--config",
                        configfile.name,
                        "_".join([argument, source]),
                        specificvalue_args,
                    ]
                    + extra_args(name, sources),
                    logger=logger,
                )

            # test default source gets defualt value
            self.assertEqual(manager.get_config(name), default_value)
            self.assertEqual(manager.get_config(name, "default"), default_value)

            # test this source gets specific value
            self.assertEqual(manager.get_config(name, source), specific_value)

            # test other sources get the default value
            for other_source in [s for s in sources if s != source]:
                self.assertEqual(manager.get_config(name, other_source), default_value)
